﻿using System;
using System.Drawing;
using System.IO;
using System.Windows.Forms;
using GASS.CUDA;
using GASS.CUDA.Direct3D;
using GASS.CUDA.Types;
using SlimDX;
using SlimDX.Direct3D10;
using SlimDX.DXGI;
using SlimDX.Windows;
using Buffer=SlimDX.Direct3D10.Buffer;
using CudaD3DDriver = GASS.CUDA.Direct3D.CUD3D10Driver;
using CudaDriver = GASS.CUDA.CUDADriver;
using Device=SlimDX.Direct3D10.Device;
using GI = SlimDX.DXGI;

namespace Example
{
    internal static class Program
    {
        private const int MeshHeight = 512;
        private const int MeshWidth = 512;
        private const int NumVertices = MeshWidth*MeshHeight;
        private const int ViewHeight = 600;
        private const int ViewWidth = 800;

        private static float anim;
        private static CUDA cuda;
        private static CUfunction cudafunc;
        private static Device device;
        private static Effect effect;
        private static InputLayout layout;
        private static RenderTargetView renderTargetView;
        private static SwapChain swapChain;
        private static EffectTechnique technique;
        private static Buffer vertexBuffer;
        private static Viewport viewport;

        [STAThread]
        private static void Main()
        {
            Application.EnableVisualStyles();
            Application.SetCompatibleTextRenderingDefault(false);
            using (var form = new Form())
            {
                form.ClientSize = new Size(ViewWidth, ViewHeight);
                Init(form);
                Application.Run(form);
                Dispose();
            }
        }

        private static void Init(Form form)
        {
            initDevice(form);
            initContent();
            initCuda();

            Application.Idle += OnApplicationIdle;
        }

        private static void initDevice(Form form)
        {
            var modeDesc = new ModeDescription();
            modeDesc.Width = ViewWidth;
            modeDesc.Height = ViewHeight;
            modeDesc.Format = Format.R8G8B8A8_UNorm;
            modeDesc.RefreshRate = new Rational(60, 1);
            modeDesc.Scaling = DisplayModeScaling.Unspecified;
            modeDesc.ScanlineOrdering = DisplayModeScanlineOrdering.Unspecified;

            var sampleDesc = new SampleDescription();
            sampleDesc.Count = 1;
            sampleDesc.Quality = 0;

            var swapDescription = new SwapChainDescription();
            swapDescription.ModeDescription = modeDesc;
            swapDescription.SampleDescription = sampleDesc;
            swapDescription.BufferCount = 1;
            swapDescription.Flags = SwapChainFlags.None;
            swapDescription.IsWindowed = true;
            swapDescription.OutputHandle = form.Handle;
            swapDescription.SwapEffect = SwapEffect.Discard;
            swapDescription.Usage = Usage.RenderTargetOutput;

            Device.CreateWithSwapChain(null, DriverType.Hardware, DeviceCreationFlags.Debug, swapDescription, out device,
                                       out swapChain);

            using (var resource = swapChain.GetBuffer<Texture2D>(0))
            {
                renderTargetView = new RenderTargetView(device, resource);
            }

            device.OutputMerger.SetTargets(renderTargetView);

            viewport = new Viewport();
            viewport.Width = ViewWidth;
            viewport.Height = ViewHeight;
            viewport.X = 0;
            viewport.Y = 0;
            viewport.MinZ = 0f;
            viewport.MaxZ = 0f;

            device.Rasterizer.SetViewports(viewport);
        }

        private static void initContent()
        {
            var bufferDesc = new BufferDescription();
            bufferDesc.BindFlags = BindFlags.VertexBuffer;
            bufferDesc.CpuAccessFlags = CpuAccessFlags.None;
            bufferDesc.OptionFlags = ResourceOptionFlags.None;
            bufferDesc.SizeInBytes = 4*sizeof (float)*NumVertices;
            bufferDesc.Usage = ResourceUsage.Default;

            vertexBuffer = new Buffer(device, bufferDesc);

            effect = Effect.FromFile(device, "effect.fx", "fx_4_0", ShaderFlags.None, EffectFlags.None, null, null);
            technique = effect.GetTechniqueByIndex(0);
            EffectPass pass = technique.GetPassByIndex(0);

            var inputElements = new[]
                                    {
                                        new InputElement("POSITION", 0, Format.R32G32B32_Float, 0, 0),
                                        new InputElement("COLOR", 0, Format.R8G8B8A8_UNorm, 12, 0)
                                    };

            layout = new InputLayout(device, inputElements, pass.Description.Signature);
        }

        private static void initCuda()
        {
            cuda = new CUDA(false);
            CUDADriver.cuInit(0);

            var cudaDevice = new CUdevice();
            CUResult res = CUDADriver.cuDeviceGet(ref cudaDevice, 0);
            CheckCudaError(res);

            CUcontext cudaContext;
            res = CUD3D10Driver.cuD3D10CtxCreate(out cudaContext, ref cudaDevice, 0, device.ComPointer);
            CheckCudaError(res);

            cuda.AttachContext(cudaContext);

            res = CUD3D10Driver.cuD3D10RegisterResource(vertexBuffer.ComPointer, CUD3D10RegisterFlags.None);
            CheckCudaError(res);
            res = CUD3D10Driver.cuD3D10ResourceSetMapFlags(vertexBuffer.ComPointer, CUD3D10MapFlags.WriteDiscard);
            CheckCudaError(res);

            CUmodule module = cuda.LoadModule(Path.Combine(Environment.CurrentDirectory, "wave.cubin"));
            cudafunc = cuda.GetModuleFunction(module, "cuda_wave");
        }

        private static void CheckCudaError(CUResult err)
        {
            if (err != CUResult.Success)
            {
                throw new Exception(err.ToString());
            }
        }

        public static void OnApplicationIdle(object sender, EventArgs e)
        {
            while (MessagePump.IsApplicationIdle)
            {
                calcWave();
                renderWave();
                anim += 0.01f;
            }
        }

        private static void calcWave()
        {
            CUResult res = CUD3D10Driver.cuD3D10MapResources(1, new[] {vertexBuffer.ComPointer});
            CheckCudaError(res);
            CUdeviceptr p;
            res = CUD3D10Driver.cuD3D10ResourceGetMappedPointer(out p, vertexBuffer.ComPointer, 0);
            CheckCudaError(res);

            cuda.SetParameter(cudafunc, 0, (uint) p.Pointer);
            cuda.SetParameter(cudafunc, IntPtr.Size, (uint) MeshWidth);
            cuda.SetParameter(cudafunc, IntPtr.Size + sizeof (int), (uint) MeshHeight);
            cuda.SetParameter(cudafunc, IntPtr.Size + sizeof (int)*2, anim);
            cuda.SetParameterSize(cudafunc, (uint) (IntPtr.Size + sizeof (int)*2 + sizeof (float)));

            cuda.SetFunctionBlockShape(cudafunc, 8, 8, 1);
            cuda.SetFunctionSharedSize(cudafunc, 0);
            cuda.Launch(cudafunc, MeshWidth/8, MeshHeight/8);

            CUD3D10Driver.cuD3D10UnmapResources(1, new[] {vertexBuffer.ComPointer});
        }

        private static void renderWave()
        {
            device.ClearRenderTargetView(renderTargetView, Color.CornflowerBlue);

            Matrix world = Matrix.Identity;
            Matrix view = Matrix.LookAtLH(new Vector3(0.0f, 3.0f, -2.0f), Vector3.Zero, Vector3.UnitY);
            Matrix proj = Matrix.PerspectiveFovLH((float) Math.PI/4.0f, ViewWidth/(float)ViewHeight, 0.01f, 10.0f);

            effect.GetVariableByName("world").AsMatrix().SetMatrix(world);
            effect.GetVariableByName("view").AsMatrix().SetMatrix(view);
            effect.GetVariableByName("projection").AsMatrix().SetMatrix(proj);

            device.InputAssembler.SetInputLayout(layout);
            device.InputAssembler.SetPrimitiveTopology(PrimitiveTopology.PointList);
            device.InputAssembler.SetVertexBuffers(0, new VertexBufferBinding(vertexBuffer, 16, 0));
            technique.GetPassByIndex(0).Apply();
            device.Draw(NumVertices, 0);

            swapChain.Present(0, PresentFlags.None);
        }

        public static void Dispose()
        {
            device.ClearState();
            vertexBuffer.Dispose();
            effect.Dispose();
            layout.Dispose();
            renderTargetView.Dispose();
            device.Dispose();
            swapChain.Dispose();
        }
    }
}